"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
  This base32 algorithm is based on Mikael Grev's MiGBase64 algorithm: http://migbase64.sourceforge.net
  which is licensed under the BSD Open Source license.
 */
var Long = require("long");
var constants_1 = require("./constants");
var shared_1 = require("../shared");
exports.REGEXP = /^[0-9a-tv-z]{2,}[0-9a-z*~$=]$/i;
var BYTE_MASK = 0xff;
var CODES = new Array(0x100);
Array.from(constants_1.SYMBOLS, shared_1.toCharCode)
    .forEach(function (charCode, i) {
    CODES[charCode] = shared_1.LONG_BYTES[i];
    var upperCode = shared_1.toCharCode(String.fromCharCode(charCode).toUpperCase());
    if (charCode !== upperCode) {
        CODES[upperCode] = CODES[charCode];
    }
});
/**
 * Douglas Crockford's base-32 symbol aliases.
 * @see http://www.crockford.com/wrmg/base32.html
 */
exports.DECODE_ALIASES = {
    "0": ["o", "O"],
    "1": ["i", "I", "l", "L"]
};
Object.keys(exports.DECODE_ALIASES).forEach(function (key) {
    var keyCode = CODES[shared_1.toCharCode(key)];
    var aliases = exports.DECODE_ALIASES[key];
    aliases.map(shared_1.toCharCode)
        .forEach(function (aliasCode) { return CODES[aliasCode] = keyCode; });
});
var CHECK_CODES = CODES.slice();
Array.from(constants_1.CHECK_EXTRAS, shared_1.toCharCode)
    .forEach(function (code, i) { return CHECK_CODES[code] = shared_1.LONG_BYTES[0x20 + i]; });
//alias the 'u' to uppercase
CHECK_CODES[shared_1.toCharCode("U")] = CHECK_CODES[shared_1.toCharCode("u")];
/**
 * Expects a string value.
 */
function decode(encoded) {
    var length = encoded.length - 1; //skip check digit
    var bytesCount = Math.trunc(length * constants_1.WORD_SIZE / constants_1.BYTE_SIZE);
    var fullWordsEnd = Math.trunc(bytesCount / constants_1.WORD_SIZE) * constants_1.WORD_SIZE;
    var result = new Uint8Array(bytesCount);
    var charPos = 0;
    var bytePos = 0;
    var checksum = 0;
    while (bytePos < fullWordsEnd) {
        var unpacked = Long.ZERO;
        for (var shift = constants_1.WORD_SHIFT_START; shift > -1; shift -= constants_1.WORD_SIZE) {
            unpacked = unpackChar(encoded, charPos++, unpacked, shift);
        }
        for (var shift = constants_1.BYTE_SHIFT_START; shift > -1; shift -= constants_1.BYTE_SIZE) {
            var byte = unpackByte(unpacked, shift);
            result[bytePos++] = byte;
            checksum += byte;
        }
    }
    // remainder
    if (bytePos < bytesCount) {
        var unpacked = Long.ZERO;
        for (var shift = constants_1.WORD_SHIFT_START; charPos < length; shift -= constants_1.WORD_SIZE) {
            unpacked = unpackChar(encoded, charPos++, unpacked, shift);
        }
        for (var shift = constants_1.BYTE_SHIFT_START; bytePos < bytesCount; shift -= constants_1.BYTE_SIZE) {
            var byte = unpackByte(unpacked, shift);
            result[bytePos++] = byte;
            checksum += byte;
        }
    }
    checksum %= constants_1.CHECK_PRIME;
    var checkDigit = encoded.charCodeAt(charPos);
    if (CHECK_CODES[checkDigit].low !== checksum) {
        throw new Error("Incorrect string -- check digits do not match. Expected: " + CHECK_CODES[checkDigit] + ", but calculated " + checksum);
    }
    return result;
}
exports.decode = decode;
function unpackChar(encoded, charPos, unpacked, shift) {
    var charCode = encoded.charCodeAt(charPos);
    var value = CODES[charCode];
    if (!value) {
        throw new Error("invalid character code: '" + charCode + "' at position " + charPos);
    }
    return unpacked.or(value.shiftLeft(shift));
}
function unpackByte(unpacked, shift) {
    return unpacked.shiftRight(shift).low & BYTE_MASK;
}
//# sourceMappingURL=decode.js.map