"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * UUID defined in IETF RFC 4122. Supports all versions of UUID.
 * @see https://tools.ietf.org/html/rfc4122.html
 */
var S = require("js.spec");
var bytes_1 = require("./bytes");
var semantic_1 = require("../semantic");
var shared_1 = require("../shared");
var uuidRegex = /^[a-fA-F\d]{8}-[a-fA-F\d]{4}-[a-fA-F\d]{4}-[a-fA-F\d]{4}-[a-fA-F\d]{12}$/;
function is16Bytes(input) {
    return bytes_1.isValidLength(input, 16, 16);
}
function matchesUUIDPattern(value) {
    return uuidRegex.test(value);
}
exports.uuidInputSpec = S.spec.or("uuid input spec", {
    "16 bytes": S.spec.and("16 bytes", bytes_1.bytesInputSpec, is16Bytes),
    "string pattern": S.spec.and("string uuid", S.spec.string, matchesUUIDPattern)
});
exports.uuidDecodingSpec = S.spec.and("uuid decoding spec", bytes_1.bytesDecodingSpec, is16Bytes);
var bytesToHex = [];
var hexToBytes = {};
for (var i = 0; i < 0x100; i++) {
    var hex = (i + 0x100).toString(16).substr(1);
    bytesToHex.push(hex);
    hexToBytes[hex] = i;
}
function forIdentifier(value) {
    return shared_1.isString(value)
        ? forStringUuid(value)
        : forBytesUuid(bytes_1.forBytesIdentifier(value));
}
var hexPos = [0, 2, 4, 6, 9, 11, 14, 16, 19, 21, 24, 26, 28, 30, 32, 34];
function forStringUuid(hex) {
    var bytes = [];
    hexPos.forEach(function (pos) {
        var hexValue = hex.substr(pos, 2);
        bytes.push(hexToBytes[hexValue]);
    });
    return createUuidLike(hex, bytes);
}
var stringPattern = Array.from("........-....-....-....-............", shared_1.toCharCode);
function forBytesUuid(bytes) {
    var chars = stringPattern.slice();
    hexPos.forEach(function (pos, i) {
        var hexValue = bytesToHex[bytes[i]];
        chars[pos++] = hexValue.charCodeAt(0);
        chars[pos] = hexValue.charCodeAt(1);
    });
    var hex = String.fromCharCode.apply(String, chars);
    return createUuidLike(hex, bytes);
}
function createUuidLike(hex, bytes) {
    return {
        bytes: bytes,
        toString: function () { return hex; },
        toJSON: function () { return hex; }
    };
}
function toDebugString(uuid) {
    return uuid.toString();
}
function encodeBytes(uuid) {
    return bytes_1.bytesCodec.encode(uuid.bytes);
}
function decodeBytes(bytes) {
    return forBytesUuid(Array.from(bytes));
}
exports.uuidCodec = {
    type: "uuid",
    typeCode: semantic_1.registerSemanticTypeCode(bytes_1.bytesCodec.typeCode, 0),
    specForIdentifier: exports.uuidInputSpec,
    specForDecoding: exports.uuidDecodingSpec,
    forIdentifier: forIdentifier,
    toDebugString: toDebugString,
    encode: encodeBytes,
    decode: decodeBytes
};
//# sourceMappingURL=uuid.js.map